"""
This module wraps code from the MIT-licensed python-mimeparse project.

The original project source code may be obtained from GitHub by visiting the
following URL:

    https://github.com/dbtsai/python-mimeparse

"""

from .mimeparse import *  # NOQA
